"""
Package 
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import pandas as pd
import param
import numpy as np
import panel as pn
import copy
from panel.viewable import Viewer
from bokeh.models.widgets.tables import NumberEditor
from scipy.stats import  mode, kurtosis, skew, entropy

from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *
from pyDendron.gui_panel.tabulator import (_cell_text_align, _cell_editors, _header_filters, array2html,
                                           _cell_formatters, _get_selection, _cell_transform, 
                                           category_html, row_date_ce, computed_filed)

class DatasetPackage(Viewer): 
    notify_package_change = param.Event()
    add_statistics = param.Boolean(default=ADD_STATISTICS, doc='Add statistics to the data')
    
    def __init__(self, dataset, parametres, editable=False, orderable=False, add_statistics=ADD_STATISTICS, **params): #param_columns, param_package, param_detrend=None, param_mean=None, editable=False, orderable=False, **params):
        super(DatasetPackage, self).__init__(**params)   
        #self.add_statistics = add_statistics
        self.param.watch(self.sync_data,  ['add_statistics'], onlychanged=True)

        self.param_column_package = parametres.column_package
        self.param_package = parametres.package
        self.param_detrend = parametres.detrend
        self.param_mean = parametres.mean
        
        dtype = self.param_column_package.get_dtype()
        columns_list = self.param_column_package.get_columns_hiddens()
        
        self._dt_data = pd.DataFrame(columns=columns_list)
        self._data = pd.DataFrame(columns=columns_list)
        self.dt_param = {}
        self.editable = editable
        self.accept_notification = True
        self.orderable = orderable
        self.bt_size = 100
        
        if self.param_detrend is not None:
            self.param_detrend.param.watch(self._sync_dt_data,  ['detrend', 'window_size', 'log'], onlychanged=True )
        
        if self.param_mean is not None:
            self.param_mean.param.watch(self._sync_dt_data,  ['biweight_mean', 'date_as_offset'], onlychanged=True)

        self.param_package.param.watch(self.sync_data, ['cambium_estimation_method', 'lambert_parameters', 'slope_resolution'], onlychanged=True)

        self.param_column_package.get_widgets().param.watch(self.sync_columns, ['value'], onlychanged=True)

        self.dataset = dataset
        self.dataset.param.watch(self.sync_dataset,  ['notify_reload', 'notify_synchronize', 'notify_packages'], onlychanged=True)

        self.wselection = pn.widgets.Select(name='Package: '+self.name, options=[], description='Select a package.')
        self.wselection.param.watch(self.sync_data,  ['value'], onlychanged=True)

        edit = _cell_editors(dtype, False)
        if self.editable:
            edit[TAG] = {'type': 'list', 'valuesLookup': True, 'autocomplete':True, 'freetext':True, 'allowEmpty':True, }

        self.wtabulator = pn.widgets.Tabulator(pd.DataFrame(columns=self.param_column_package.get_columns_hiddens()),
                                    hidden_columns=self.param_column_package.get_hiddens(), 
                                    text_align=_cell_text_align(dtype),
                                    editors=edit, 
                                    on_edit=self.on_edit,
                                    header_filters=_header_filters(dtype), 
                                    formatters=_cell_formatters(dtype),
                                    frozen_columns=[ICON, KEYCODE, 'index'], 
                                    show_index=False,
                                    pagination= 'local',
                                    page_size=100, #100000,
                                    selectable='checkbox', 
                                    sizing_mode='stretch_width',
                                    max_height=400,
                                    min_height=300,
                                    height_policy='max',
                                    row_content = self.get_row_content,
                                    ) 

        self.panel_tabulator = pn.Card(self.wtabulator, margin=(5, 0), collapsed=True, 
                                       sizing_mode='stretch_width',  
                                       title='Data '+self.name, collapsible=True,
                                       )
        
        stylesheet = 'p {padding: 0px; margin: 0px;}'
        self.dt_info = pn.pane.Alert('Detrend data is empty set', margin=(0, 0, 5, 5), align=('start', 'end'), stylesheets=[stylesheet])
        
        self._layout = pn.Column(pn.Row(self.wselection, self.dt_info), self.panel_tabulator)

    def set_package_name(self, package_name):
        """
        Set the package name in the GUI panel.

        Args:
            package_name (str): The name of the package to set.

        Returns:
            None
        """
        self.wselection.value = package_name

    def get_package_name(self):
        """
        Returns the name of the package selected in the GUI panel.
        
        Returns:
            str: The name of the selected package.
        """
        return self.wselection.value

    def get_row_content(self, series):
        """
        Returns a view of a datavalue.

        Parameters:
        - series: The series containing the data values.

        Returns:
        - pn.Tabs: A panel containing the data values in a tabular format.

        Raises:
        - Exception: If there is an error retrieving the data values.
        """        
        try:
            if isinstance(series[DATA_VALUES], np.ndarray):
                #perror(f'get_row_content {series.name}', series.info())
                lst = []
                lst.append((RAW, array2html(series[DATA_VALUES])))
                if SLOPE in series:
                    lst.append((SLOPE, array2html(series[SLOPE])))
                if self._dt_data is not None:
                    dt_type = self._dt_data.at[series.name, DATA_TYPE]
                    if dt_type != RAW:
                        lst.append((dt_type, array2html(self._dt_data.at[series.name, DATA_VALUES])))
                return pn.Tabs(*lst)
        except Exception as inst:
            return pn.pane.Markdown('Data error.')
        return pn.pane.Markdown('Data is missing.')

    def __panel__(self):
        return self._layout

    def sync_columns(self, event):
        """
        Set the hidden columns in the tabulator widget.

        Parameters:
        - event: The event object triggered by the column selector.

        Returns:
        None
        """
        self.param_column_package.update_tabulator(self.wtabulator)

    def sync_dataset(self, event):
        """
        Synchronizes the dataset with the GUI panel.

        This method updates the package selection options.

        Parameters:
            event (object): The event object triggered by the sync action.

        Returns:
            None
        """
        if not self.accept_notification:
            return
        lst = self.dataset.package_keys()
        columns_list = self.param_column_package.get_columns_hiddens()

        self._data = pd.DataFrame(columns=columns_list)
        self.wtabulator.value = pd.DataFrame(columns=columns_list)
        self._dt_data = pd.DataFrame(columns=columns_list)
        self.dt_param = {}
        self.wselection.options = ['None'] + lst
        if self.wselection.value not in self.wselection.options:
            self.wselection.value = 'None' 
        else:
            self.sync_data(event)
    
    def sync_data(self, event):
        """
        Synchronizes the data in the GUI panel with the dataset package.

        Args:
            event: The event object triggered by the synchronization action.

        Returns:
            None
        """
        if not self.accept_notification:
            return
        try:
            self._layout.loading = True
            columns_list = self.param_column_package.get_columns_hiddens()
            self._data = pd.DataFrame(columns=columns_list)
            package_name = self.get_package_name()
            if package_name != 'None':
                self._data = self.dataset.get_package_components(package_name, self.param_package.slope_resolution, self.param_package).reset_index()
                computed_filed(self._data, icon=True, date=True)
                self._data.reset_index(inplace=True)
                self._data = self._data.sort_values(by=KEYCODE)
        except Exception as inst:
            self._data = pd.DataFrame(columns=self.param_column_package.get_columns_hiddens())
        finally:
            #perror(f'sync_data:', self.name, self.add_statistics)
            if self.add_statistics:
                self.statistics()
            self.param_column_package.update_tabulator(self.wtabulator, self._data, self.add_statistics) 
            self._sync_dt_data(event)
            self._layout.loading = False

    def _sync_dt_data(self, event):
        """
        Synchronizes the detrended data with the current dataset.

        This method performs the detrending operation on the dataset based on the specified parameters.
        It updates the detrended data and the detrend information accordingly.

        Args:
            event: The event triggering the synchronization.

        Returns:
            None
        """
        def get_dt_param():
            dt_param = {}
            if self.param_detrend is not None:
                dt_param[DETREND] = self.param_detrend.detrend
                dt_param[DETREND_WSIZE] = self.param_detrend.window_size
                dt_param[DETREND_LOG] = self.param_detrend.log
                dt_param[MEANDATE_AS_OFFSET] = self.param_mean.date_as_offset
                dt_param[BIWEIGHT_MEAN] = self.param_mean.biweight_mean
            return dt_param
        
        try:
            self._layout.loading = True
            tmp_data = self.get_data()
            
            tmp_dt_data = tmp_data
            ids = tmp_data[ID_CHILD].unique().tolist()
            if len(ids) <= 0:
                tmp_dt_data = None
                tmp_data = None
                self.dt_info.object = 'Detrend data is empty set'
                self.dt_info.alert_type = 'warning'
            elif (self.param_detrend is  None) or (self.param_detrend.detrend == RAW):
                self.dt_info.object = 'Detrend data is raw data. '
                self.dt_info.alert_type = 'primary'
            else:
                if len(ids) != len(tmp_data[ID_CHILD]):
                    logger.warning(f'Duplicate series in package {self.name}')
                tmp_dt_data = tmp_data[[ID_CHILD, ID_PARENT, OFFSET, CAMBIUM_LOWER, CAMBIUM_ESTIMATED, CAMBIUM_UPPER, SLOPE]]
                res = self.dataset.detrend(ids, self.param_detrend.detrend, self.param_detrend.window_size, 
                                                    self.param_detrend.log, self.param_mean.date_as_offset, 
                                                    self.param_mean.biweight_mean, self.param_package.slope_resolution, self.param_mean.min_elements)      
                tmp_dt_data = tmp_dt_data.join(res, on=ID_CHILD, how='left')
                self.dt_info.alert_type = 'info'
                c = f'Detrend data is {self.get_data_type()} '
                c += ', '.join([f'{index}: {valeur}' for index, valeur in tmp_dt_data[CATEGORY].value_counts().items()]) +'.'
                self.dt_info.object = c
                
                if tmp_dt_data[INCONSISTENT].any():
                    self.dt_info.object += ' one or more series is inconsistent.'
                    self.dt_info.alert_type='warning'
                else:
                    self.dt_info.alert_type='primary'
        except Exception as inst:
            self.dt_info.object = f'Detrend data is {RAW} data'
            logger.error(f'_sync_dt_data: {inst}', exc_info=True)
        finally:
            self.dt_param = get_dt_param()
            self._dt_data = tmp_dt_data
            self.accept_notification = False
            self.param.trigger('notify_package_change')
            self.accept_notification = True
            self._layout.loading = False
    
    def on_cambium_estimation(self, event):
        """
        Handle the event when cambium estimation is triggered.

        Args:
            event: The event object representing the cambium estimation event.

        Returns:
            None
        """
        try:
            self._layout.loading = True
            edit = self.wtabulator.editors
            if self.param_package.cambium_estimation_method == 'user values':
                edit[CAMBIUM_LOWER] = NumberEditor(step=1)
                edit[CAMBIUM_ESTIMATED] = NumberEditor(step=1)
                edit[CAMBIUM_UPPER] = NumberEditor(step=1)
            else:
                edit[CAMBIUM_LOWER] = None
                edit[CAMBIUM_ESTIMATED] = None
                edit[CAMBIUM_UPPER] = None
            self.wtabulator.editors = edit

            self.do_cambium_estimation(self._data)
            self.wtabulator.value = _cell_transform(self._data)
            self.param.trigger('notify_package_change')
        except Exception as inst:
            logger.error(f'sync_data: {inst}', exc_info=True)
        finally:
            self._layout.loading = False

    def get_data(self):
        """
        Retrieves mean and tree from the wtabulator and returns it.

        If the wtabulator is empty, an empty DataFrame is returned.

        Returns:
            pandas.DataFrame: The retrieved data.
        """
        return self._data

    @property
    def data(self):
        """
        Returns the selected data from the tabulator widget.

        If there is no selection, returns the entire dataset.

        Returns:
            pandas.DataFrame: The selected data.
        """
        if (self._data is None) or len(self._data) == 0:
            return None
        d = _get_selection(self.wtabulator)
        
        ids = self.wtabulator.current_view.index if len(d) <= 0 else d.index
        df = self._data.loc[ids,:]     
        return df.loc[df[CATEGORY].isin([MEAN, MEASURE]),:]

    def get_data_type(self):
        """
        Returns the detrend data as a string.

        Returns:
            str: The detrend data type as a string.
        """
        if (self.param_detrend is  None) or (self.param_detrend.detrend == RAW):
            return f'{RAW}'
        if self.param_detrend.log and (self.param_detrend.detrend != BP73):
            return f'log({self.param_detrend.detrend}), ws: {self.param_detrend.window_size}. '
        
        return f'{self.param_detrend.detrend}, ws: {self.param_detrend.window_size}. '
    
    @property
    def dt_data(self):
        """
        Returns a subset of the detrend dataset based on the selected indices.

        If no indices are selected, the entire dataset is returned.

        Returns:
            pandas.DataFrame: Subset of the dataset.
        """
        if (self._dt_data is None) or len(self._dt_data) == 0:
            return None
        d = _get_selection(self.wtabulator)
        
        ids = self.wtabulator.current_view.index if len(d) <= 0 else d.index
        df = self._dt_data.loc[ids,:]     
        return df.loc[df[CATEGORY].isin([MEAN, MEASURE]),:]

    def _apply_log(self, array):
        x = np.log(array)
        x[np.isinf(x)] = np.nan
        return x

    @property
    def log_data(self):
        """
        Apply logarithm to the data values in the dataset.

        Returns:
            DataFrame: A copy of the dataset with logarithm applied to the data values.
        """
        
        if (self._data is None) or len(self._data) == 0:
            return None
        d = _get_selection(self.wtabulator)

        df = copy.deepcopy(self.data)
        df[DATA_VALUES] = df[DATA_VALUES].apply(self._apply_log)
        ids = self.wtabulator.current_view.index if len(d) <= 0 else d.index
        df = df.loc[ids,:]     
        return df.loc[df[CATEGORY].isin([MEAN, MEASURE]),:]

    def get_data_values_id(self, id):
        def get_data(df, id):
            if df is None:
                return None
            return df.loc[df[ID_CHILD] == id, DATA_VALUES].iloc[0]
        
        data = get_data(self.data, id)
        
        dt_data = get_data(self.dt_data, id)
        log_data = get_data(self.log_data, id) 
        
        return data, dt_data, log_data

    def on_edit(self, event):
        """
        Handle the event when a cell is edited.

        Args:
            event: The event object representing the edit event.
        """
        try:
            self.accept_notification = False
            self._layout.loading = True
            col = event.column
            row = self.wtabulator.value.iloc[event.row]
            new = event.value
            id_parent, id_child = row[ID_PARENT], row[ID_CHILD]
            self._data.at[row.name, col] = new
            self._dt_data.at[row.name, col] = new
            if col == TAG:
                current_package = self.get_package_name()
                self.dataset.edit_sequence(id_child, col, new, notify=True)
            else:
                self.param_package.cambium_estimation_method = 'user values'
        except Exception as inst:
            self.wtabulator.patch({event.column: [(event.row, event.old)]})
            logger.error(f'on_edit: {inst}', exc_info=True)
        finally:            
            self._layout.loading = False    
            self.param.trigger('notify_package_change')
            self.accept_notification = True
    
    def save(self):
        """
        Saves the package using the current tabulator value.

        Note: This method does not trigger the 'notify_package_change' event.

        Returns:
            None
        """
        save_package(self._data, self.get_package_name(), self.dataset)

    def statistics(self):
        """
        Calculate statistics for the dataset.

        Args:
            columns (list, optional): List of columns to include in the statistics. Defaults to [ID, KEYCODE].
            stat_columns (list, optional): List of additional statistics columns to include. Defaults to [DATA_NAN].
            data (pandas.DataFrame, optional): Data to calculate statistics on. Defaults to None.

        Returns:
            pandas.DataFrame: DataFrame containing the calculated statistics.

        """
        if self._data is not None:
            # Filtrer les lignes où DATA_LENGTH est non nul et supérieur à 0
            valid_rows = self._data[DATA_LENGTH].notna() & (self._data[DATA_LENGTH] > 0)
            if valid_rows.sum() > 0:
                values = self._data.loc[valid_rows, DATA_VALUES]                
                self._data.loc[valid_rows, STAT_MEAN] = values.apply(lambda x: np.nanmean(x))
                self._data.loc[valid_rows, STAT_MEDIAN] = values.apply(lambda x: np.nanmedian(x))
                self._data.loc[valid_rows, STAT_MODE] = values.apply(lambda x: mode(x, nan_policy='omit')[0])
                self._data.loc[valid_rows, STAT_STD] = values.apply(lambda x: np.nanstd(x))
                self._data.loc[valid_rows, STAT_VAR] = values.apply(lambda x: np.nanvar(x))
                self._data.loc[valid_rows, STAT_MIN] = values.apply(lambda x: np.nanmin(x))
                self._data.loc[valid_rows, STAT_MAX] = values.apply(lambda x: np.nanmax(x))
                self._data.loc[valid_rows, STAT_PERC25] = values.apply(lambda x: np.nanpercentile(x, 25))
                self._data.loc[valid_rows, STAT_PERC50] = values.apply(lambda x: np.nanpercentile(x, 50))
                self._data.loc[valid_rows, STAT_PERC75] = values.apply(lambda x: np.nanpercentile(x, 75))
                self._data.loc[valid_rows, STAT_SUM] = values.apply(lambda x: np.nansum(x))
                self._data.loc[valid_rows, STAT_KURTOSIS] = values.apply(lambda x: kurtosis(x, nan_policy='omit'))
                self._data.loc[valid_rows, STAT_SKEWNESS] = values.apply(lambda x: skew(x, nan_policy='omit'))
                self._data.loc[valid_rows, STAT_ENTROPY] = values.apply(lambda x: entropy(x, nan_policy='omit'))

def save_package(dataframe, package_name, dataset):
    
    def get_missing_key(df, key):
        mask = df[key].isna()
        return  mask, np.sum(mask)

    def get_missing_values(df):
        mask = (df[CATEGORY] != SET) & df[DATA_VALUES].isna()
        return mask, np.sum(mask)
    
    if package_name == '':
        logger.warning(f'Selection name is empty')
    else:
        df = dataframe
        paires = list(set(zip(df[ID_PARENT], df[ID_CHILD])))   
               
        mask, smask = get_missing_key(df, DATE_BEGIN)
        if smask > 0:       
            logger.warning(f'Some {DATE_BEGIN} is missing')
        mask, smask  = get_missing_key(df, OFFSET)
        if smask > 0:       
            logger.warning(f'Some {OFFSET} is missing')
        mask, smask = get_missing_values(df)
        if smask > 0:       
            logger.warning(f'Some {DATA_VALUES} is missing, remove them')
            df = df.loc[~mask]
        if len(df) != 0:      
            dataset.set_package(package_name, paires)
            logger.info(f'Save selection {package_name}')
        else:
            logger.warning(f'Selection is empty, delete package.')
            #dataset.delete_package(package_name)


                
            

        







"""
    pyDendron panel
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
 
import os
import getpass
import argparse
import sys

import panel as pn
from pathlib import Path
import json
from bokeh.server.contexts import BokehSessionContext
import bokeh
from datetime import datetime
import platform
import shutil

from pyDendron.dataname import ADD_STATISTICS
from pyDendron.dataset import Dataset
from pyDendron.gui_panel.sidebar import ParamColumns, ParamDetrend, ParamMean, ParamPackage, ParamColumnPackage
from pyDendron.gui_panel.dataset_selector import DatasetSelector
from pyDendron.gui_panel.dataset_treeview import DatasetTreeview
from pyDendron.gui_panel.dataset_package import DatasetPackage
from pyDendron.gui_panel.dataset_package_builder import DatasetPackageBuilder
from pyDendron.gui_panel.dataset_package_editor import DatasetPackageEditor
from pyDendron.gui_panel.tools_panel import ToolsPanel
from pyDendron.gui_panel.ploter_panel import PloterPanel
from pyDendron.gui_panel.debug_panel import DebugPanel
from pyDendron.gui_panel.crossdating_panel import CrossDatingPanel
from pyDendron.app_logger import logger, perror, catch_bokeh_log, open_mail_client, LOG_FILENAME, check_version, __version__, catch_tornado_log


def set_extension():
    """
    Set the required extensions for the pyDendron panel.

    This function enables the necessary extensions for the pyDendron panel to function properly.
    It enables the 'throttled', 'notifications', 'tabulator', and 'loading_spinner' extensions.
    It also sets the loading indicator to True and disables the console output.

    Parameters:
        None

    Returns:
        None
    """
    def version(request):
        return __version__
    
    pn.extension(session_key_func=version) # set a session key, use for --reuse-sessions
    pn.extension('tabulator', 'filedropper', throttled=True, notifications=True, loading_spinner='dots')
    pn.param.ParamMethod.loading_indicator = True
    pn.extension(
        disconnect_notification="""Server Connected Closed <br /> <button class="btn btn-outline-light" onclick="location.reload();">Click To Reconnect</button> """,
        ready_notification="""connected."""
    )

class Parameters:
    """
    Class representing the parameters for pyDendron.
    """

    def __init__(self, app, cfg_path):
        self.app = app
        self.detrend = None
        self.mean = None 
        self.columns = None 
        self.column_package = None
        self.package = None
        self.cfg_path = cfg_path
        os.makedirs(os.path.dirname(cfg_path), exist_ok=True)
        self.cfg_filename = cfg_path / Path('pyDendron.cfg.json')
        self.bt_save_parameters = pn.widgets.Button(name='Save parameters', icon="settings", button_type='default', align=('end', 'center'))
        self.bt_save_parameters.on_click(self.on_save_param)
        self.load()

    def get_widgets(self):
            """
            Returns the widget for saving parameters.

            Returns:
                bt_save_parameters (Widget): The widget for saving parameters.
            """
            return self.bt_save_parameters
    
    def save(self):
        """
        Save the parameters to a file.

        This method serializes the parameters of various components and saves them to a JSON file.

        Returns:
            None
        """
        with open(self.cfg_filename, 'w') as f:
            data = {
                'version': __version__,
                'detrend' : self.detrend.param.serialize_parameters(),
                'mean' : self.mean.param.serialize_parameters(),
                'columns' : self.columns.get_columns(),
                'column_package': self.column_package.get_columns(),
                'package' : self.package.param.serialize_parameters(),
            }
            json.dump(data, f)
        self.app.ploter.dump_cfg()
        self.app.crossdating.dump_cfg()
        self.app.dataset_selector.dump_cfg()
        self.app.tools.dump_cfg()
    
    def on_save_param(self, event):
        """
        Event handler for saving parameters.

        This method saves the current parameters and displays a notification
        indicating that the parameters have been saved.

        Args:
            event: The event object associated with the save action.

        Returns:
            None
        """
        self.save()
        logger.info('Parameters saved')
    
    def load(self):
        """
        Load the parameters from a file.

        Args:
            cfg_filename (str): The path to the file.

        Raises:
            Exception: If there is an error loading the file.
        """
        def default_parameters():
            self.detrend = ParamDetrend()
            self.mean = ParamMean()
            self.columns = ParamColumns()
            self.column_package = ParamColumnPackage()
            self.package = ParamPackage()
        
        import json
        try:
            if check_version(self.cfg_filename):
                with open(self.cfg_filename, 'r') as f:
                    data = json.load(f)
                    self.detrend = ParamDetrend(**ParamDetrend.param.deserialize_parameters(data['detrend']))
                    self.mean = ParamMean(**ParamMean.param.deserialize_parameters(data['mean']))
                    self.columns = ParamColumns(selections=data['columns'])
                    self.column_package = ParamColumnPackage(selections=data['column_package'])
                    self.package = ParamPackage(**ParamPackage.param.deserialize_parameters(data['package']))
            else:
                default_parameters()
        except Exception as inst:
            logger.warning(f'1 - Error in {self.cfg_filename}, ignore cfg files. {inst}')
            default_parameters()


class SidebarWidget:
    """
    A class representing a sidebar widget.

    Attributes:
        sidebar (object): The sidebar object.
        objects (dict): A dictionary to store the sidebar objects.
        hiddens (dict): A dictionary to store the paramters to hide depending of ative tab in main panel.
    """

    def __init__(self, app):
        # Add to sidebar
        self.sidebar = app.template.sidebar
        self.objects = {}
        self.hiddens = {}
        self.append_title('## Datasets')
        self.append(app.dataset_selector, 'dataset')
        self.append_title('## Parameters')
        self.append(app.parameters.columns, 'param_columns')
        self.append(app.parameters.column_package, 'param_column_package')
        self.append(app.parameters.package, 'param_package')
        self.append(app.parameters.detrend, 'param_detrend') 
        self.append(app.parameters.mean, 'param_mean')
        self.append(app.treeview, 'treeview')
        self.append(app.ploter, 'ploter')
        self.append(app.crossdating, 'crossdating')
        self.append(app.tools, 'tools')
        
        self.init_hiddens()
        self.active_parameters(app.treeview.name)
        
    def append(self, pane, name):
        """
        Append an parameter object to the sidebar.

        Args:
            pane (Viewer): The pane to be appended.
            name (str): The name of the object.
        """
        card = pane.get_sidebar()
        self.objects[name] = card
        self.sidebar.append(card)

    def append_title(self, title):
        """
        Append a title to the sidebar.

        Args:
            title (str): The title to be appended.
            name (str): The name of the title.
        """
        self.sidebar.append(title)

    def init_hiddens(self):
        """
        Initialize the parameter objects to hide for a given the title of the tab.
        """
        self.hiddens['Treeview'] = ['ploter', 'crossdating', 'tools', 'param_detrend', 'param_package','param_column_package']
        self.hiddens['Package'] = ['treeview', 'ploter', 'crossdating', 'tools', 'param_detrend', 'param_mean', 'param_columns']
        self.hiddens['Ploter'] = ['treeview', 'crossdating', 'tools', 'param_columns']
        self.hiddens['Crossdating'] = ['treeview', 'ploter', 'tools', 'param_columns']
        self.hiddens['Tools'] = ['treeview', 'ploter', 'crossdating', 'tools', 'param_package', 
                                 'param_detrend', 'param_mean', 'param_columns']
        self.hiddens['Debug'] = ['treeview', 'ploter', 'crossdating', 'tools', 'param_colums', 'param_package', 
                                 'param_detrend', 'param_mean', 'param_columns', 'param_column_package']

    def active_parameters(self, name):
        """
        Activate the parameters based on the given name of the active tab.

        Args:
            name (str): The name to determine which parameters to activate.
        """
        lst = self.hiddens[name]
        for param_name, param in self.objects.items():
            param.visible = False if param_name in lst else True

class DendronApp:
    """
    The main application class for pyDendron.
    
    Args:
        dendron_info (object): An object containing information about pyDendron.
    """
    def __init__(self, www, session_context):
        filters = ['*.p', '*.json', '*.xlsx']

        self.www = www
        self.session_context = session_context
        self.username = getpass.getuser() if pn.state.user is None else pn.state.user
        
        if f'context {self.username}' not in pn.state.cache:        
            self.dendron_path = None
            self.dataset_path = None
            self.cfg_path = None
            self.tmp_path = None
            
            self.set_directories()
            self.parameters = Parameters(self, self.cfg_path)
            self.template = self.get_template()       
            self.dataset = Dataset(username=self.username)
            self.dataset_selector = DatasetSelector(self.dataset, template=self.template, path=self.dataset_path, filters=filters, cfg_path=self.cfg_path )        
            
            self.treeview = DatasetTreeview(self.dataset, self.parameters, name='Treeview')
            self.tab_package = self.get_tab_package()
            self.ploter = PloterPanel(self.dataset, self.parameters, cfg_path=self.cfg_path, name='Ploter')
            self.crossdating = CrossDatingPanel(self.dataset, self.parameters, self.cfg_path, name='Crossdating')
            self.tools = ToolsPanel(self.dataset_selector, self.dataset, self.parameters, self, tmp_path=self.tmp_path, cfg_path=self.cfg_path, filters=filters, name='Tools')
            self.debug = DebugPanel(self.dataset, self.parameters, self, name='Debug')
            self.tabs = pn.Tabs(self.treeview, self.tab_package, self.ploter, self.crossdating, self.tools, ('Debug', self.debug),
                    dynamic=False, margin=0, styles={'padding' : '0px', 'font-size': '18px'}, sizing_mode='stretch_both',)

            self.tabs.param.watch(self.on_activate, ['active'], onlychanged=True)
            self.template.main.append(self.tabs)
            self.sidebar_widget = SidebarWidget(self)
            

        else:
            self.get_cache()

    def set_cache(self):
        #self.ploter.reset()

        d = {}
        print('--> Set cache')
        print('--> session_context (set_cache)', self.session_context)
        pn.state.cache[f'context {self.username}'] = d
        #d['session_context'] = self.session_context
        d['dendron_path'] = self.dendron_path 
        d['dataset_path'] = self.dataset_path
        d['cfg_path'] = self.cfg_path 
        d['tmp_path'] = self.tmp_path
        d['parameters'] = self.parameters
        d['dataset_selector'] = self.dataset_selector
        d['dataset'] = self.dataset
        #d['tabs'] = self.tabs
        d['treeview'] = self.treeview
        d['tab_package'] = self.tab_package
        #d['ploter'] = self.ploter
        #d['crossdating'] = self.crossdating
        d['tools'] = self.tools
        d['debug'] = self.debug
        #d['template'] = self.template
        #d['sidebar_widget'] = self.sidebar_widget
        d['ploter.dataset_package'] = self.ploter.dataset_package
        d['crossdating.dataset_package'] = self.crossdating.dataset_package
        d['crossdating.master_dataset_package'] = self.crossdating.master_dataset_package
        d['tabs.active'] = self.tabs.active
        d['crossdating.cross_dating.results'] = self.crossdating.cross_dating.results

    def get_cache(self):
        def set_package(package, data):
            package.set_package_name(data.get_package_name())
            package.sync_dataset(None)
        
        print('--> Get cache')
        if f'context {self.username}' in pn.state.cache:
            d = pn.state.cache[f'context {self.username}']

            self.dendron_path = d['dendron_path']
            self.dataset_path = d['dataset_path']
            self.cfg_path = d['cfg_path']
            self.tmp_path = d['tmp_path']
            self.parameters = d['parameters']
            
            self.template = self.get_template()       

            #self.sidebar_widget = d['sidebar_widget']
            self.dataset_selector = d['dataset_selector']
            self.dataset = d['dataset']
            self.treeview = d['treeview']
            self.tab_package = d['tab_package']
            self.ploter = PloterPanel(self.dataset, self.parameters, cfg_path=self.cfg_path, name='Ploter')
            self.crossdating = CrossDatingPanel(self.dataset, self.parameters, self.cfg_path, name='Crossdating')
            self.tools = d['tools']
            self.debug = d['debug']
            self.tabs = pn.Tabs(self.treeview, self.tab_package, self.ploter, self.crossdating, self.tools, ('Debug', self.debug),
                    dynamic=False, margin=0, styles={'padding' : '0px', 'font-size': '18px'}, sizing_mode='stretch_both',)

            self.tabs.param.watch(self.on_activate, ['active'], onlychanged=True)
            self.template.main.append(self.tabs)
            self.sidebar_widget = SidebarWidget(self)
            
            set_package(self.ploter.dataset_package, d['ploter.dataset_package'])
            set_package(self.crossdating.dataset_package, d['crossdating.dataset_package'])
            set_package(self.crossdating.master_dataset_package, d['crossdating.master_dataset_package'])
            self.crossdating.cross_dating.results = d['crossdating.cross_dating.results']
            self.crossdating.on_tabs(None)
            self.tabs.active = d['tabs.active']
            
        
    def set_directories(self):
        self.dendron_path = Path(os.path.expanduser("~")) / 'pyDendron'
        self.dataset_path = self.dendron_path / 'dataset' / Path(self.username)
        self.cfg_path = self.dendron_path / 'cfg' / Path(self.username)
        self.tmp_path = self.dendron_path / 'tmp' / Path(self.username)

        os.makedirs(self.dataset_path, exist_ok=True)
        os.makedirs(self.cfg_path, exist_ok=True) 
        os.makedirs(self.tmp_path, exist_ok=True) 
        
        dataset_test = self.dataset_path / 'dataset_test.p'
        dataset_test_src = os.path.join(os.path.dirname(__file__), 'data', 'dataset_test.p')
        if (not dataset_test.is_file()) and (Path(dataset_test_src).is_file()):
            shutil.copy(dataset_test_src, self.dataset_path)
        self.print_info()

    def print_info(self):
            logger.info(f'Mode: {self.www}')
            logger.info(f'username: {self.username}')
            logger.info(f'dendron_path: {self.dendron_path}')
            logger.info(f'dataset_path: {self.dataset_path}')
            logger.info(f'cfg_path: {self.cfg_path}')
            logger.info(f'tmp_path: {self.tmp_path}')

    def get_tab_package(self):
        """
        Get the tab package.
        
        Returns:
            pn.Tabs: The tab panel containing the package builder and package editor.
        """
        package_editor = DatasetPackageEditor(self.dataset, self.parameters, name='Editor')
        package_builder = DatasetPackageBuilder(self.dataset, self.parameters, self.treeview, name='Builder')
        tab_package = pn.Tabs(package_builder, package_editor, name='Package',
                                styles={'font-size': '16px'})
        return tab_package

    def bt_notification(self):
        """
        Create a button to clear notifications.
        
        Returns:
            pn.widgets.Button: The button to clear notifications.
        """
        def on_rm_notification(event):
            pn.state.notifications.clear()
            
        rm_notification = pn.widgets.Button(name='Clear notifications', icon="trash", button_type='default', align=('end', 'center'))
        rm_notification.on_click(on_rm_notification)
        return rm_notification
    
    def bt_gps(self):
        """
        Create a button to open the GPS site.
        
        Returns:
            pn.pane.Markdown: The button to open the GPS site.
        """
        # GPS 
        return pn.pane.Markdown("""<svg xmlns="http://www.w3.org/2000/svg" width="18" height="18" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="1" stroke-linecap="round" stroke-linejoin="round" class="icon icon-tabler icons-tabler-outline icon-tabler-map">
        <path stroke="none" d="M0 0h24v24H0z" fill="none" />
        <path d="M3 7l6 -3l6 3l6 -3v13l-6 3l-6 -3l-6 3v-13" />
        <path d="M9 4v13" />
        <path d="M15 7v13" />
        </svg>
        <a href="https://nominatim.openstreetmap.org" target="_blank">Open GPS site</a>""")

    def get_logout(self):
        return pn.pane.Markdown("""<svg  xmlns="http://www.w3.org/2000/svg"  width="18"  height="18"  viewBox="0 0 24 24"  fill="none"  stroke="currentColor"  stroke-width="1"  stroke-linecap="round"  stroke-linejoin="round"  class="icon icon-tabler icons-tabler-outline icon-tabler-logout">
            <path stroke="none" d="M0 0h24v24H0z" fill="none"/><path d="M14 8v-2a2 2 0 0 0 -2 -2h-7a2 2 0 0 0 -2 2v12a2 2 0 0 0 2 2h7a2 2 0 0 0 2 -2v-2" />
            <path d="M9 12h12l-3 -3" /><path d="M18 15l3 -3" /></svg>
            <a href="/logout" >Logout</a>""")    


    def bt_send_log(self):
        def on_send_log(event):
            
            panel_version = pn.__version__
            bokeh_version = bokeh.__version__
            date_today = datetime.now()
            os_info = platform.system() + ' ' + platform.release()
            user_name = pn.state.user #getpass.getuser()
            
            agent = pn.state.headers["User-Agent"] if "User-Agent" in pn.state.headers else "Unknown"
            
            info_string = f"""
            pyDendron: {__version__}
            Date: {date_today}
            OS: {os_info}
            User: {user_name}
            Agent: {agent}
            Panel: {panel_version}
            Bokeh: {bokeh_version}"""

            info_string += '\n'+'-'*80 + '\n'
            with open(LOG_FILENAME, 'r') as file:
                info_string += file.read()
                
            open_mail_client(
                to='pyDendron@univ-lemans.fr',
                subject=f'pyDendron error {user_name}',
                body=info_string
            )
        
        bt_log = pn.widgets.Button(name='Send log', icon="send", button_type='default', align=('end', 'center'))
        bt_log.on_click(on_send_log)
        return bt_log

    def get_version(self):
        return pn.pane.Markdown(f"version: {__version__} user: {self.username}")

    def get_template(self):
        """
        Get the template for the application.
                
        Returns:
            pn.template.FastListTemplate: The template for the application.
        """
        logo = os.path.join(os.path.dirname(__file__), 'data', 'trees.svg')
        logo_blue = os.path.join(os.path.dirname(__file__), 'data', 'trees_blue.svg')
        # site_url='https://git-lium.univ-lemans.fr/Meignier/pyDendron/wikis',
        template = pn.template.FastListTemplate(title='pyDendron', logo=logo, favicon=logo_blue, meta_author='LIUM, Le Mans Université', 
                                                meta_keywords='Dendrochronology', main_layout=None)
        template.header.objects = [self.get_version(), self.bt_notification(), self.parameters.get_widgets(), 
                                    self.bt_send_log(), self.bt_gps()]
        if self.www:
            template.header.objects.append( self.get_logout())
        template.modal.append(pn.Column())
        return template
            
    def on_activate(self, event):
        """
        Event handler for tab activation.
        
        Args:
            event (object): The event object.
        """
        self.sidebar_widget.active_parameters(self.tabs.objects[event.obj.active].name)
        
app = None
          
def cb_auto_save():
    """
    Automatically save the dataset.
    """
    msg = 'save'
    if (app is not None) and app.dataset_selector.save_auto:
        msg += ' dataset'
        app.dataset.dump()
    if (app is not None) and app.dataset_selector.save_cfg:
        msg += ' parameters'
        app.parameters.on_save_param(None)
    if msg != 'save':
        logger.info(msg)

def cb_disconnection(app):
    
    #if pn in globals() and (logger is not None):
    try:
        logger.info(f'On disconnection {pn.state.user} {app.username}.')
        if (app is not None):
            logger.info(f' Save context: start')
            app.set_cache()
            app.dataset.dump()
            app.parameters.on_save_param(None)
    except Exception as e:
        print('*** Warning: Context not save***')
        print('*** error', e)

def cb_ready_notification():
    """
    Display a notification when the application is ready.
    """
    pn.state.notifications.append('pyDendron is ready.')

def cb_connection(app):
    pass

#def session_created(session_context: BokehSessionContext):
#    print('--> session created', session_context)
    
def session_destroyed(session_context: BokehSessionContext):
    print('--> session destroyed', session_context)

app = None
if __name__.startswith('bokeh_app'):
    #pn.state.on_session_created(session_created)
    print('--> session_context (main)', pn.state.curdoc.session_context)
    pn.state.on_session_destroyed(session_destroyed)
    
    parser = argparse.ArgumentParser(description="pydenron: A dendromean tool for tree-ring data analysis.")
    parser.add_argument('--www', action='store_true', help='A flag to enable www mode')
    args = parser.parse_args()
    www = args.www

    logger.info('Get paths')
    set_extension()
    logger.info('Create application ')
    app = DendronApp(www, pn.state.curdoc.session_context)
    logger.info('End of initialisation')
    app.template.servable()
    # Callback: autosave 
    logger.info('add_periodic_callback ')
    pn.state.add_periodic_callback(cb_auto_save, 1000*60*5) 
    catch_bokeh_log(cb_connection, cb_disconnection, app)
    #catch_tornado_log(app)
    


    



"""
    Import  file 
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"


from pathlib import Path
import pickle
import io
import re
import os
import zipfile
import shutil
import pandas as pd


from pyDendron.dataname import *
from pyDendron.tools.location import reverse_geocode, get_elevation
from pyDendron.dataset import Dataset
from pyDendron.app_logger import logger, perror


class IO:
    def __init__(self, encoding='utf-8', source='unk', places='reverse_places.p', get_place=False, get_altitude=False, base_directory='./'):
        self.sequences = []
        self.components = []
        self.id = 1
        self.base_directory = base_directory
        self.source = source
        self.places = {}
        self.elevations = {}
        self.encoding = encoding
        self.get_place = get_place
        self.get_altitude = get_altitude
        self.places_fn = places
        
        if get_place:
            self._read_places()

    def _read_places(self):
        if (self.places_fn is not None) and Path(self.places_fn).exists():
            with open(self.places_fn , 'rb') as fic:
                self.places, self.elevations = pickle.load(fic)

    def _write_places(self):
        with open(self.places_fn , 'wb') as fic:
            pickle.dump([self.places, self.elevations], fic)

    def _get_location(self, meta):
        meta[SITE_COUNTRY] = meta[SITE_STATE] = meta[SITE_DISTRICT] = meta[SITE_TOWN] = meta[SITE_ZIP] = ''
        if self.get_place:
            meta[SITE_COUNTRY], meta[SITE_STATE], meta[SITE_DISTRICT], meta[SITE_TOWN], meta[SITE_ZIP] = reverse_geocode(meta[SITE_LATITUDE], meta[SITE_LONGITUDE], self.places)
        if self.get_altitude:
            meta[SITE_ELEVATION] = get_elevation(meta[SITE_LATITUDE], meta[SITE_LONGITUDE], self.elevations)

    def init(self, keycode_parent):
        parent_id = self.next_id()
        self.sequences.append({ID: WORKSHOP, KEYCODE: 'Workshop', CATEGORY: SET})
        self.sequences.append({ID: TRASH, KEYCODE: 'Trash', CATEGORY: SET})
        self.sequences.append({ID: CLIPBOARD, KEYCODE: 'Clipboard', CATEGORY: SET})
        self.sequences.append({ID: ROOT, KEYCODE: 'Dataset', CATEGORY: SET})
        self.sequences.append({ID: parent_id, KEYCODE: keycode_parent, CATEGORY: SET})
        self.components.append({ID_PARENT: ROOT, ID_CHILD:parent_id, OFFSET: pd.NA})

        return parent_id

    def _readline(self, fd):
        line = fd.readline().strip()
        while line == '':
            line = fd.readline().strip()
        return line
    
    def next_id(self):
        self.id += 1
        return self.id

    def read_sequences(self, id_parent, lines):
        pass
    
    def read_files(self, directory):
        lst = []
        for ext in self.file_extension:
            lst += [x for x in directory.rglob('*'+ext) if x.is_file()]
        
        perror('lst:', lst)
        concatenated_bytes = b''
        for filename in lst:
                with open(filename, 'br') as file:
                    concatenated_bytes += file.read()
        return concatenated_bytes

    def zip_to_buffer(self, buffer, keycode_parent):
        output_directory = Path(self.base_directory) / keycode_parent
        os.makedirs(output_directory, exist_ok=True)
        
        output_file = output_directory / Path(keycode_parent).with_suffix(".zip")
        with open(output_file, 'wb') as f:
            f.write(buffer)  

        zip = zipfile.ZipFile(output_file)
        zip.extractall(output_directory)
        buffer = self.read_files(output_directory)
        
        shutil.rmtree(output_directory)
        return buffer

    def read_buffer(self, keycode_parent, buffer, mine_type=None):
        parent_id = self.init(keycode_parent)
        if  mine_type == 'application/zip':
            buffer = self.zip_to_buffer(buffer, keycode_parent)
        elif mine_type == 'text/plain':
            buffer = buffer.encode(self.encoding)
            
        iobuffer = io.BytesIO(buffer)
        text = iobuffer.getvalue().decode(self.encoding, errors='ignore')
        text = re.sub(r'\r\n?|\n', '\n', text)
        lines = text.split('\n')
        #lines = iobuffer.readlines()
        lines = [line+'\n' for line in lines]
        #print(lines)
        self.read_sequences(parent_id, lines)
        self._write_places()
        
        return Dataset(sequences=self.sequences, components=self.components)
    
    def read(self, filename):
        with open(Path(filename), 'rt', encoding=self.encoding, errors='ignore', newline=None) as fd:
            keycode_parent = Path(filename).stem
            parent_id = self.init(keycode_parent)
            lines = fd.readlines()
            #print('lines:',lines)
            self.read_sequences(parent_id, lines)

        self._write_places()

        return Dataset(sequences=self.sequences, components=self.components)

    def write(self, data, means, filename):
        pass

    def write_package(self, dataset, data, filename):
        id_samples = data.loc[data[CATEGORY] == MEASURE, ID_CHILD].to_list()
        means = {}
        for _, row in data[data[CATEGORY] == MEAN].iterrows():
            tree = dataset.get_descendants(row[ID_CHILD])
            samples = {dataset.sequences.at[node.id, KEYCODE]: offset for node, offset in tree.descendants[MEASURE].items() if node.id not in id_samples}
            means[row[ID_CHILD]] = samples
            
        self.write_file(data, means, filename)
            
    
    
    



"""
    Import tridas file 
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import glob
#import xml.etree.ElementTree as ET
import logging
import copy
import re
from lxml import etree as ET
import numpy as np
import pandas as pd
import pickle
import zipfile
import os
import shutil

from pyDendron.tools.location import reverse_geocode, get_elevation
from pyDendron.dataset import Dataset
from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *
from pathlib import Path

class IOTridas:
    file_extension = ['tridas.xml']
    
    def __init__(self, glob_str=None, base_directory= './', places='reverse_places.p', source=None, get_place=True, get_altitude=True):
        self.glob_str = '*'+IOTridas.file_extension[0] if glob_str is None else glob_str
        self.base_directory = Path(base_directory)
        self.uri = None
        self.source = source
        self.laboratories = 'empty'
        self.id = 0
        self.sequences = []
        self.components = []
        self.tree = None
        self.id_means = None
        self.meta_means = None
        self.id_projects = None
        self.meta_projects = None
        self.get_place = get_place
        self.get_altitude = get_altitude
        self.namespaces = {}
        self.identifier = {}
        self.places = {}
        self.elevations = {}
        self.places_fn = places
        self.dataset = None
            
    def _read_places(self):
        if (self.places_fn is not None) and Path(self.places_fn).exists():
            #print('read places')
            with open(self.places_fn , 'rb') as fic:
                self.places, self.elevations = pickle.load(fic)

    def _write_places(self):
        with open(self.places_fn , 'wb') as fic:
            pickle.dump([self.places, self.elevations], fic)
        
    def add_component(self, id_parent, id_child, offset=np.nan):
        self.components.append({ID_PARENT: id_parent, ID_CHILD: id_child, OFFSET: offset})
    
    #def add_raw_indices(self, id, values):
    #    self.indices.append({ID: id, 'key': Indices.RAW, 'values': values, 'count': len(values)})

    def _init_metadata(self, category, category_tridas, parent_meta=None):
        meta = {}
        if parent_meta is not None:
            meta = copy.deepcopy(parent_meta)
        #meta[SOURCE] = self.source
        #meta[LABs] = self.laboratories
        #meta[URI] = self.uri
        meta[ID] = self.id
        meta[CATEGORY] = category
        meta[SUBCATEGORY] = category_tridas
        #meta[SAPWOOD] = pd.NA
        #meta[PITH] = pd.NA
        #meta[CAMBIUM] = pd.NA
        #meta[BARK] = pd.NA
        
        self.id += 1
        return meta, self.id -1

    def _new_metadata(self, node, parent_meta, category=SET, category_tridas='unk'):
            id_parent = parent_meta[ID]
            meta, id = self._init_metadata(category, category_tridas, parent_meta)
            self._add_keycode(meta, node)
            self.sequences.append(meta)
            self.add_component(id_parent, id)
            return meta, id
    
    def _set_laboratories(self, project):
        self.laboratories = '' 
        for lab in project.findall('tridas:laboratory/tridas:name', self.namespaces):
            if 'acronym' in lab.attrib:
                self.laboratories += lab.attrib['acronym'] + ', '
            else:
                self.laboratories += lab.text + ', '
        self.laboratories = re.sub(', $', '', self.laboratories)

    def _add_keycode(self, meta, node):
        meta[KEYCODE] = node.find('tridas:title', self.namespaces).text
        self._set_identifier(meta, node)

    def _set_identifier(self, meta, node):
        key = self._get_identifier(node)
        self.identifier[key] = meta

    def _get_identifier(self, node):
        _id = node.find('tridas:identifier', self.namespaces)
        if _id is not None:
            domain = self.laboratories
            if 'domain' in _id.attrib:
                domain = _id.attrib['domain']
            if _id.text is None:
                return domain + '// warning empty text'
            return domain + '//' + _id.text
        return None

    def _add_taxon(self, meta, node):
        if node.find('tridas:taxon', self.namespaces) is not None:
            meta[SPECIES] = node.find('tridas:taxon', self.namespaces).text
        
    def _add_location(self, meta, node):
        d = {}
        if node.find('tridas:location/tridas:locationGeometry/gml:Point/gml:pos', self.namespaces) is not None:
            (longitude, latitude) = node.find('tridas:location/tridas:locationGeometry/gml:Point/gml:pos', self.namespaces).text.split(' ')
            latitude, longitude = float(latitude), float(longitude)
            meta[SITE_LATITUDE] = latitude
            meta[SITE_LONGITUDE] = longitude
            if not((latitude < 72) and (latitude > 36) and (longitude > -9.5) and (longitude < 28)):
                 if (longitude < 72) and (longitude > 36) and (latitude > -9.5) and (latitude < 28):
                     latitude, longitude = longitude, latitude
                     #logger.warning(f'location not in europe, invert longitude and lattitude: {latitude}, {longitude}')
                 else:
                     logger.warning(f'location not in europe: {latitude}, {longitude}')
            meta[SITE_COUNTRY] = meta[SITE_STATE] = meta[SITE_DISTRICT] = meta[SITE_TOWN] = meta[SITE_ZIP] = ''
            meta[SITE_ELEVATION] = np.nan
            if self.get_place:
                _, __, ___, ____, _____, meta[SITE_CODE] = reverse_geocode(meta[SITE_LATITUDE], meta[SITE_LONGITUDE], self.places)
#                meta[SITE_COUNTRY], meta[SITE_STATE], meta[SITE_DISTRICT], meta[SITE_TOWN], meta[SITE_ZIP] = reverse_geocode(latitude, longitude, self.places)
            if self.get_altitude:
                meta[SITE_ELEVATION] = get_elevation(meta[SITE_LATITUDE], meta[SITE_LONGITUDE], self.elevations)
#               meta[SITE_ELEVATION] = get_elevation(latitude, longitude, self.elevations)
        
    def presence_of(self, node, key, warning=True):
        keycode = node.find('tridas:title', self.namespaces).text

        if node.find(key, self.namespaces) is not None:
            if warning:
                logger.warning(f'{key} id present but not read in {keycode}')
            return True
        return False
    
    def _presence_of_wood_attribut(self, node):
        if 'presence' in node.attrib:
            if node.attrib['presence'] in ['incomplete', 'complete', 'present']:
                return True
        return False
       
    def _read_wood_completeness(self, meta, node):
        
        info_wood = node.find('tridas:woodCompleteness', self.namespaces)
        if info_wood is not None:
            if info_wood.find('tridas:ringCount', self.namespaces) is not None:
                meta[DATA_LENGTH] = int(info_wood.find('tridas:ringCount', self.namespaces).text)
            if info_wood.find('tridas:pith', self.namespaces) is not None:
                meta[PITH] = self._presence_of_wood_attribut(info_wood.find('tridas:pith', self.namespaces))
            if info_wood.find('tridas:nrOfUnmeasuredInnerRings', self.namespaces) is not None:
                meta[PITH_ESTIMATED] = int(info_wood.find('tridas:nrOfUnmeasuredInnerRings', self.namespaces).text)
                #heartwood = self._presence_of_wood_attribut(info_wood.find('tridas:heartwood', ns))
            #if info_wood.find('tridas:sapwood', self.namespaces) is not None:
            #    meta[SAPWOOD] = self._presence_of_wood_attribut(info_wood.find('tridas:sapwood', self.namespaces))
            if info_wood.find('tridas:sapwood/tridas:nrOfSapwoodRings', self.namespaces) is not None:
                meta[SAPWOOD] = meta[DATA_LENGTH] - int(info_wood.find('tridas:sapwood/tridas:nrOfSapwoodRings', self.namespaces).text)
            if info_wood.find('tridas:sapwood/tridas:lastRingUnderBark', self.namespaces) is not None:
                meta[CAMBIUM_SEASON] = self._presence_of_wood_attribut(info_wood.find('tridas:sapwood/tridas:lastRingUnderBark', self.namespaces))
            #if info_wood.find('tridas:sapwood/tridas:missingSapwoodRingsToBark', self.namespaces) is not None:
            #    meta[CAMBIUM_OPTIMUM] = meta[DATA_LENGTH] + int(info_wood.find('tridas:sapwood/tridas:missingSapwoodRingsToBark', self.namespaces).text)
            #if info_wood.find('tridas:sapwood/tridas:missingSapwoodRingsToBarkFoundation', self.namespaces) is not None:
            #    meta[CAMBIUM_METHOD] = info_wood.find('tridas:sapwood/tridas:missingSapwoodRingsToBarkFoundation', self.namespaces).text
            if info_wood.find('tridas:bark', self.namespaces) is not None:
                meta[BARK] = self._presence_of_wood_attribut(info_wood.find('tridas:bark', self.namespaces))
    
    def _interpretation(self, meta, node):
        def date(node):
            c = 1.0
            d = 0
            if node.attrib['suffix'] in ['BC', 'BCE']:
                c = -1
                d = 1
            elif node.attrib['suffix'] == 'BP':
                #logger.debug('date BP !')
                c = -1
                d = 1950
            date = int(node.text) * c + d
            if 'certainty' in node.attrib:
                return date, node.attrib['certainty']
            return date, 'unknown'
        
        if node.find('tridas:interpretation/tridas:firstYear', self.namespaces) is not None:
            y = node.find('tridas:interpretation/tridas:firstYear', self.namespaces)
            meta[DATE_BEGIN], meta[SYNC] = date(y)
        if node.find('tridas:interpretation/tridas:lastYear', self.namespaces) is not None:
            y = node.find('tridas:interpretation/tridas:lastYear', self.namespaces)   
            meta[DATE_END], meta[SYNC] = date(y)
    
    # ------------------------------------
    def _readall(self):
        self._read_places()

        self.meta_projects, self.id_projects = self._init_metadata(SET, 'projects')
        self.meta_projects[KEYCODE] = 'projects'
        self.sequences.append(self.meta_projects)
        
        self.meta_means, self.id_means = self._init_metadata(SET, 'means')
        self.meta_means[KEYCODE] = 'means'
        self.sequences.append(self.meta_means)
        
        self._read(is_projetcs=True) # read all project without derivedSeries
        self._read(is_projetcs=False) # read all derivedSeries

        self._write_places()
       
    def _read(self, is_projetcs=True):
        self.namespaces = {}
        lst = list(self.base_directory.rglob(self.glob_str))
        print(f'self.base_directory: {self.base_directory}')
        print(f'self.glob_str: {self.glob_str}')
        print(f'lst: {lst}')
        #lst = [x for x in glob.glob(self.glob_str)]
        for i, self.filename in enumerate(lst):
            if i % 100 == 0:
                logger.info(f'{np.round((i/len(lst))*100, 2)}, {self.filename}, {is_projetcs}')
                #logger.debug('_write_places')
                self._write_places()
            
            self.uri = Path(self.filename).resolve().as_uri()
            tree = ET.parse(self.filename)
            root = tree.getroot()

            for key, value in root.nsmap.items():
                if key is None:
                    key = 'default'
                self.namespaces[key] = value
            
            if root.tag == '{'+self.namespaces["tridas"]+'}project':
                if is_projetcs:
                    self._read_project(root, self.meta_projects.copy())
                else:
                    self._read_mean(root)
            
    def _read_project(self, project, parent_meta):
        #identifier, createdTimestamp, lastModifiedTimestamp, comments, type_[], description, laboratory[], 
        # category, investigator,period, requestDate, commissioner, reference[], research[], genericField[]
        #self.presence_of(project, 'tridas:file')
        meta, _ = self._new_metadata(project, parent_meta, SET, PROJECT)
        self._set_laboratories(project)
        meta[PROJECT] = meta[KEYCODE]
        #print('_read_project')
        for object in project.findall('tridas:object', self.namespaces):
            self._read_object(object, meta)
        
    def _read_mean(self, project):
        self._set_laboratories(project)

        for derived_series in project.findall('tridas:derivedSeries', self.namespaces):
            self._read_derived_series(derived_series, project)
            
    def _read_object(self, object, parent_meta):
        #identifier, createdTimestamp, lastModifiedTimestamp, comments, type_, description,
        #, creator, owner, coverage, genericField
        
        #self.presence_of(object, 'tridas:file')
        self.presence_of(object, 'tridas:linkSeries')
        
        meta, id = self._new_metadata(object, parent_meta, SET, 'object')
        self._add_location(meta, object)
        
        for object_ in object.findall('tridas:object', self.namespaces):
            self._read_object(object_, meta)
        
        elements = object.findall('tridas:element', self.namespaces)
        if elements:
            for element in elements:
                self._read_element(element, meta, len(elements))
        
    def _read_element(self, element, parent_meta, nb):
        #identifier, createdTimestamp, lastModifiedTimestamp, comments, type_, description,
        #shape, dimensions, authenticity, processing, marks, altitude,  slope, soil, bedrock, genericField,
        
        self.presence_of(element, 'tridas:linkSeries')

        if nb == -1 and (element.find('tridas:keycode', self.namespaces).text == parent_meta[KEYCODE]):
            meta = parent_meta
            meta[SUBCATEGORY] += ', element'
        else:
            meta, id = self._new_metadata(element, parent_meta, SET, category_tridas='element')

        self._add_location(meta, element)
        self._add_taxon(meta, element)

        samples = element.findall('tridas:sample', self.namespaces)
        if samples is not None:
            for sample in samples:
                self._read_sample(sample, meta, len(samples)) 
    
    def _read_sample(self, sample, parent_meta, nb):
        #identifier, createdTimestamp, lastModifiedTimestamp, comments, type_, description,
        #samplingDate, offset, state, knots, genericField,

        #self.presence_of(sample, 'tridas:file')

        if nb == -1 and (sample.find('tridas:keycode', self.namespaces).text == parent_meta[KEYCODE]):
            meta = parent_meta
            meta[SUBCATEGORY] += ', sample'
        else:
            meta, id = self._new_metadata(sample, parent_meta, SET, category_tridas=MEASURE)

        radius = sample.findall('tridas:radius', self.namespaces)
        if radius is not None:
            for radius_ in radius:
                self._read_radius(radius_, meta, len(radius) ) 
    
    def _read_radius(self, radius, parent_meta, nb):
        #identifier, createdTimestamp, lastModifiedTimestamp, comments, azimuth, genericField

        if nb == -1 and (radius.find('tridas:keycode', self.namespaces).text == parent_meta[KEYCODE]):
            meta = parent_meta
            meta[SUBCATEGORY] += ', radius'
        else:
            meta, id = self._new_metadata(radius, parent_meta, SET, category_tridas='radius')

        self._read_wood_completeness(meta, radius)

        measurement_series = radius.findall('tridas:measurementSeries', self.namespaces)
        if measurement_series is not None:
            for ms in measurement_series:
                self._read_measurement_series(ms, meta, len(measurement_series)) 

    def _read_measurement_series(self, measurement_series, parent_meta, nb):
        # keycode, identifier, createdTimestamp, lastModifiedTimestamp, comments, id, measuringDate, derivationDat,  
        # analyst, dendrochronologist, measuringMethod, type_, objective, standardizingMethod, author, version, 
        # interpretationUnsolved,  genericField
        
        self.presence_of(measurement_series, 'tridas:linkSeries')
  
        if nb == -1 and (measurement_series.find('tridas:keycode', self.namespaces).text == parent_meta[KEYCODE]):
            meta = parent_meta
            meta[CATEGORY] = MEASURE
            meta[SUBCATEGORY] += ', mSeries'
        else:
            meta, id = self._new_metadata(measurement_series, parent_meta, MEASURE, category_tridas='mSeries')

        self._read_wood_completeness(meta, measurement_series)
        self._add_location(meta, measurement_series)
        self._interpretation(meta, measurement_series)
      
        for v in measurement_series.findall('tridas:values', self.namespaces):
            vect, _ = self._read_values(v, meta[ID]) 
            meta[DATA_VALUES] = vect
            meta[DATA_LENGTH] = len(vect)
            meta[DATA_TYPE] = 'raw'

    def _read_derived_series(self, derived_series, project):
        #identifier, createdTimestamp, lastModifiedTimestamp, comments, id, measuringDate, derivationDate,  
        #analyst, dendrochronologist, measuringMethod, type_, objective, standardizingMethod, author, version, 
        #  interpretationUnsolved,  genericField=None, 
        if derived_series.find('tridas:keycode', self.namespaces) is None:
            #logger.warning(f'keycode is missing in derived_series')
            return
        
        keycode = derived_series.find('tridas:keycode', self.namespaces).text
        
        links = derived_series.findall('tridas:linkSeries/tridas:series', self.namespaces)     
        on_set = False   
        if len(links) == 1 :
            id_domain = self._get_identifier(links[0])
            if (id_domain in self.identifier) and (self.identifier[id_domain][CATEGORY] == SET):
                meta = self.identifier[id_domain]
                meta[CATEGORY] = SET
                id = meta[ID]
                on_set = True
        if not on_set :   
            for link in links:
                id_domain = self._get_identifier(link)
                if id_domain not in self.identifier:
                    logger.warning(f'Remove derived_series: {id_domain} is missing in {keycode}')
                    return
            
            meta, id = self._init_metadata(MEAN, 'dSeries')
            self._add_keycode(meta, derived_series)
            self.sequences.append(meta)
            meta[PROJECT] = project.find('tridas:keycode', self.namespaces).text
            self.add_component(self.id_means, id)
            for link in links:
                self.add_component(id, self.identifier[self._get_identifier(link)][ID])

        self._read_wood_completeness(meta, derived_series)
        self._add_location(meta, derived_series)
        self._interpretation(meta, derived_series)

        for v in derived_series.findall('tridas:values', self.namespaces):
            vect, weigths = self._read_values(v, id)
            meta[DATA_VALUES] = vect
            meta[DATA_WEIGHTS] = weigths
            meta[DATA_LENGTH] = len(vect)
            meta[DATA_TYPE] = 'raw'

    def _read_values(self, values, id):
        vect = []
        weigths = []
        for value in values.findall('tridas:value', self.namespaces):
            v = float(value.attrib['value'])
            if (v == 9999) or (v == 0):
                v = np.nan
            vect.append(v)
            if 'count' in value.attrib:
                weigths.append(float(value.attrib['count']))
        
        vect = [np.nan if (x == 999) or (x <= 0) else x for x in vect[:-1]]

        return np.array(vect), np.array(weigths)
    
    def to_dataset(self, keycode_parent='Dataset', root_id=ROOT, trash_keycode='Trash', 
                   clipboard_keycode='Clipboard', workshop_keycode='Workshop'):
        self._readall()
        self.dataset = Dataset(sequences=self.sequences, components=self.components, save_auto=False)

        self.dataset.new_root(keycode_parent, root_id)
        self.dataset.new_clipboard(clipboard_keycode)
        self.dataset.new_trash(trash_keycode)
        self.dataset.new_workshop(workshop_keycode)
        return self.dataset
    
    def merge_level(self):        
        # merge level with only one son
        def iterate(node):
            sep = '\\'
            for child in node.children:
                if child.category != MEASURE:
                    iterate(child)
            if  (len(node.children) == 1) and (node.parent is not None) : 
                node.parent.children = node.children
                id_parent = node.parent.id
                id = node.id
                id_child = node.children[0].id
                s = self.dataset.sequences
                if s.at[id_child, KEYCODE] != s.at[id, KEYCODE]:
                    s.at[id_child, KEYCODE] = s.at[id, KEYCODE] + sep + s.at[id_child, KEYCODE]
                cols = list(set(s.columns) - set([SUBCATEGORY, KEYCODE, CATEGORY, DATA_LENGTH, DATA_VALUES, DATA_WEIGHTS]))
                for col in cols:
                    if pd.isna(s.at[id_child, col]):
                        s.at[id_child, KEYCODE] = s.at[id, KEYCODE]
                    elif pd.notna(s.at[id, col]):
                        if s.at[id_child, col] != s.at[id, col]:
                            logger.info(f'\t copy [{id_child} <-- {id}, {col}] {s.at[id_child, col]} <-- {s.at[id, col]}')
                                             
                p = self.dataset.components.at[(id, id_child), OFFSET]
                self.dataset.soft_drop(id)
                self.dataset.components.at[(id_parent, id_child), OFFSET] = p
                                
        tree = self.dataset.get_descendants(self.dataset.get_roots())
        iterate(tree)
        return self.dataset

        
    def read_buffer(self, keycode_parent, buffer, mine_type=None):

        #if mine_type != 'application/zip':
        #    raise ValueError(f'Unsupported mine_type {mine_type}')
        
        output_directory = Path(self.base_directory) / keycode_parent
        os.makedirs(output_directory, exist_ok=True)
        self.base_directory = output_directory
        

        if mine_type == 'application/zip':
            output_file = output_directory / Path(keycode_parent).with_suffix(".zip")
            with open(output_file, 'wb') as f:
                f.write(buffer)  
            zip = zipfile.ZipFile(output_file)
            zip.extractall(output_directory)
        else:
            output_file = output_directory / Path(keycode_parent).with_suffix(".tridas.xml")
            with open(output_file, 'w') as f:
                    f.write(buffer)  
        
        dataset = self.to_dataset(keycode_parent=keycode_parent) 
        shutil.rmtree(output_directory)
        return dataset  



"""
Parametre classes for the sidebar
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import os
import param
import copy
import panel as pn
from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *
from pyDendron.crossdating import CrossDating
from pyDendron.detrend import detrend_types
from pyDendron.gui_panel.tabulator import _cell_transform


class ParamColumns:
    def __init__(self, selections=[ICON, KEYCODE, DATE_BEGIN, DATE_END, OFFSET], dtype=dtype_view, title='Columns'):
        self.dtype = dtype
        options_list = selections + list(set(dtype.keys()) - set(selections)) 
        
        self.columns = self.columns = pn.widgets.MultiChoice(name='Columns', value=selections, options=options_list, 
                                     sizing_mode='stretch_width', search_option_limit=25, description='Select columns to display in table.')
        self.title = title
        self.move_field = pn.widgets.Select(name='Move field', options=selections, value='', sizing_mode='stretch_width') 
        self.columns.param.watch(self.on_columns, ['value'], onlychanged=True, queued=True)
        
        self.bt_up = pn.widgets.Button(name='Up', icon="arrow-up", align=('start', 'end'))
        self.bt_up.on_click(self.on_up)
        
        self.bt_down = pn.widgets.Button(name='Down', icon="arrow-down", align=('start', 'end'))
        self.bt_down.on_click(self.on_down)

        self.card = pn.Card(self.columns, pn.Row(self.move_field, self.bt_up, self.bt_down), title=self.title, sizing_mode='stretch_width', margin=(5, 0), collapsed=True)
        
    def on_columns(self, event):
        sv = self.move_field.value
        self.move_field.options = event.new
        self.move_field.value = sv if sv in event.new else ''
    
    def on_up(self, event):
        move_field = self.move_field.value
        columns = copy.deepcopy(self.columns.value)
        if move_field is not None:
            i = columns.index(move_field)
            if i > 0:
                columns[i], columns[i-1] = columns[i-1], columns[i]
                self.columns.options = columns + list(set(self.columns.options) - set(columns))                
                self.columns.value = columns
                self.move_field.value = move_field                
    
    def on_down(self, event):
        move_field = self.move_field.value
        columns = copy.deepcopy(self.columns.value)
        if move_field is not None:
            i = columns.index(move_field)
            if i < len(columns)-1:
                columns[i], columns[i+1] = columns[i+1], columns[i]
                self.columns.options = columns + list(set(self.columns.options) - set(columns))                
                self.columns.value = columns
                self.move_field.value = move_field
                   
    def handle_columns(self, change):
        self.columns_hide.value = change.new

    def get_dtype(self):
        return self.dtype

    def get_hiddens(self):
        return list(set(self.get_options()) - set(self.get_columns())) 

    def get_columns_hiddens(self):
        return self.get_columns() + self.get_hiddens()

    def get_columns(self):
        return  self.columns.value
    
    def get_options(self):
        return  self.columns.options

    def get_widgets(self):
        return self.columns
        
    def get_sidebar(self):    
        return self.card  
    
    def update_tabulator(self, wtabulator, data=None, add_statistics=False):
        wtabulator.hidden_columns=self.get_hiddens() 
        if wtabulator.value is not None:
            if data is None:
                data = wtabulator.value
            col = self.get_columns_hiddens() 
            if add_statistics:
                col += stat_dtype_dict.keys()
            col = [c for c in col if c in data.columns]
            wtabulator.value = _cell_transform(data[col])         

class ParamColumnPackage(ParamColumns):
    def __init__(self, selections=[ICON, KEYCODE, DATE_BEGIN, DATE_END, OFFSET], dtype=dtype_package, title='Package Columns'):
        super().__init__(selections, dtype, title)

class ParamColumnCrossDating(ParamColumns):
    def __init__(self, selections=list(set(CrossDating.COLS.keys())- set([ID, ID_MASTER])), dtype=CrossDating.COLS, title='Statistic Columns'):
        super().__init__(selections, dtype, title)


class ParamMean(param.Parameterized):
    #num_threads = param.Integer(default=1, bounds=(1, os.cpu_count()), step=1, doc='Number of threads.')
    biweight_mean =  param.Boolean(False, doc='Biweight mean method for computing the mean.')
    date_as_offset =  param.Boolean(True, doc='Dates are used as offsets for computing the mean.')
    check_mean_as_measure = param.Boolean(False, doc='Check if Mean is considered as a measure.')
    min_elements = param.Integer(default=4, bounds=(1, 30), step=1, doc='Minimal number of elements for computing the mean.')
    def get_sidebar(self):    
        return pn.Card(pn.Param(self, show_name=False), title='Mean', sizing_mode='stretch_width', margin=(5, 0), collapsed=True)  

class ParamDetrend(param.Parameterized):
    #num_threads = param.Integer(default=1, bounds=(1, 10), step=1, doc='Number of threads.')
    detrend = param.Selector(objects=detrend_types, doc='Detrend method apply to the series.')
    window_size = param.Integer(default=5, bounds=(3, 15), step=2, doc='Size of the sliding window for detrending.')
    log = param.Boolean(True, doc='Perform a logarithm transformation at the end of detrending.')
            
    def __init__(self, **params):
        super(ParamDetrend,self).__init__(**params)
    
    def get_sidebar(self):
        return pn.Card(pn.Param(self, show_name=False), title='Detrend', sizing_mode='stretch_width', margin=(5, 0), collapsed=True)  

class ParamPackage(param.Parameterized):
    #show_data = param.Boolean(False, doc='Show data in selection view')
    cambium_estimation_method = param.Selector(default='Lambert', objects=['Lambert', 'log-log', 'user values'], doc='Method for cambium estimation.')
    lambert_parameters = param.Range(default=(12,23), bounds=(0, 60), step=1, doc='Lambert estimator range.')
    #pith_estimation = param.Boolean(False, doc='Draw pith estimation')
    slope_resolution = param.Integer(default=0, bounds=(0, 10), step=1, doc='Minimal GLK resolution')
            
    def __init__(self, **params):
        super(ParamPackage,self).__init__(**params)
    
    def get_sidebar(self):
        return pn.Card(pn.Param(self, show_name=False), title='Package', sizing_mode='stretch_width', margin=(5, 0), collapsed=True)  



"""
Crossdating tools
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import logging
import pandas as pd
import matplotlib
matplotlib.use('Agg')

import panel as pn
import json
import param
import time
import numpy as np
import warnings

from panel.viewable import Viewer
from bokeh.io import export_svgs, export_svg
from bokeh.plotting import figure
from bokeh.models import ColumnDataSource
from bokeh.palettes import Category10
from bokeh.events import DoubleTap
from pathlib import Path
from io import BytesIO

from pyDendron.app_logger import logger, perror, check_version, __version__
from pyDendron.dataname import *
from pyDendron.crossdating import CrossDating, COLS, MapValueError, r
from pyDendron.gui_panel.dataset_package import DatasetPackage
from pyDendron.gui_panel.run import RunThread
from pyDendron.gui_panel.sidebar import ParamColumnCrossDating
from pyDendron.gui_panel.tabulator import (_cell_text_align, _cell_editors, _header_filters_crossdating,
                                           _cell_formatters, _get_selection,
                                             get_download_folder, unique_filename)

class CrossDatingPanel(Viewer):
    """A panel for cross-dating analysis.

    This panel provides functionality for cross-dating analysis, including filtering, visualization, and exporting of results.

    Parameters:
    - dataset (Dataset): The dataset for cross-dating analysis.
    - parameters (Parameters): The parameters for cross-dating analysis.
    - cfg_path (str): The path to the configuration file.

    Attributes:
    - inner (bool): Flag indicating whether inner cross-dating is enabled.
    - param_array (ParamArray): Parameters related to the array crossdating.
    - param_filter (ParamFilter): Parameters related to filtering of results.
    - cfg_file (str): The path to the configuration file.
    - results (None or DataFrame): The computed results of cross-dating analysis.
    - pmatrix (None or Matplotlib Figure): The matrix plot of cross-dating analysis.
    - pstem (None or Bokeh Figure): The stem plot of cross-dating analysis.
    - gplot (None or Bokeh Figure): The graph plot of cross-dating analysis.
    - dataset_package (DatasetPackage): The dataset package for cross-dating analysis.
    - master_dataset_package (DatasetPackage): The master dataset package for cross-dating analysis.
    - cross_dating (CrossDating): The cross-dating object.
    - wtabulator (Tabulator): The tabulator widget for displaying results.
    - plot (None or Plot): The plot widget for displaying the selected result.
    - wrun (Row): The row widget containing the run buttons.
    - warray (Column): The column widget containing the array-related widgets.
    - wheat_matrix (Matplotlib): The Matplotlib widget for displaying the matrix plot.
    - wstem (Bokeh): The Bokeh widget for displaying the stem plot.
    - wzstem (Bokeh): The Bokeh widget for displaying the zoomed stem plot.
    - whist (Column): The column widget for displaying the histogram plot.
    - wmap (Bokeh): The Bokeh widget for displaying the map plot.
    - wgraph (Bokeh): The Bokeh widget for displaying the graph plot.
    - tabs (Tabs): The tabs widget for switching between different views.

    Methods:
    - get_tabulator(): Returns the Tabulator widget for displaying results.
    - _sync_data(event): Synchronizes the data when the dataset package changes.
    - get_plot(row): Returns the plot widget for the selected result.

    """

    inner =  param.Boolean(True, doc='Inner crossdation')

    class ParamArray(param.Parameterized):
        table_height = param.Integer(400, bounds=(100, 1000), step=10, doc='Maximum height of the Treeview tabulator.')
        max_results = param.Integer(default=500000, allow_None=True, bounds=(100000, 5000000), step=100000, doc='Maximum number of results displayed.')
        group_by = param.Selector(default=None, objects=[None, KEYCODE, KEYCODE_MASTER, KEYCODE+'/'+KEYCODE_MASTER, DATE_END_ESTIMATED, SYNC, T_RANK, Z_RANK, D_RANK], doc='Group scores by column.')
#        columns = param.ListSelector(default=list(set(list(CrossDating.COLS.keys()))- set([ID, ID_MASTER])), 
#                                objects=list(CrossDating.COLS.keys()),
#                                doc='Displayed crossdating columns.')
    param_array = ParamArray(name='Array')
    
    class ParamFilter(param.Parameterized):
        score = param.Selector(default=T_SCORE, objects=[CORR, T_SCORE, GLK, Z_SCORE, DISTANCE], doc='Score applyed with threshold.')
        filter_threshold = param.Boolean(False, doc='Apply filter using score threshold.')
        threshold = param.Number(default=0, allow_None=True, bounds=(None, None), step=0.5, doc='Keep results upper the threshold.')
        filter_max_rank = param.Boolean(True, doc='Apply filter using rank value.')
        max_rank = param.Integer(default=10, allow_None=True, bounds=(1, None), step=1, doc='Keep top ranking results based on score field.')
        filter_sync = param.Boolean(False, doc=f'Apply filter on {SYNC}.')
        sync = param.Boolean(True, doc='Apply filter using rank value.')
        no_dates_in_the_future = param.Boolean(False, doc="Apply a filter on dates greater than today's date for start and end dates.")
    param_filter = ParamFilter()


    def __init__(self, dataset, parameters, cfg_path, **params):
        super(CrossDatingPanel, self).__init__(**params)   
        self.cfg_file = cfg_path / Path(f'{self.__class__.__name__}.cfg.json')
        self.param_array_columns = ParamColumnCrossDating()

        bt_size = 120
        self.pmatrix = None
        self.pstem = None
        self.gplot = None
        self.row_results = None
        self.dataset_package = DatasetPackage(dataset, parameters, name='hyp')
        self.master_dataset_package = DatasetPackage(dataset, parameters,name='master')
        self.master_dataset_package._layout.visible = False

        self.dataset_package.param.watch(self._sync_data, ['notify_package_change'], onlychanged=True)
        self.master_dataset_package.param.watch(self._sync_data, ['notify_package_change'], onlychanged=True)

        self.cross_dating = self.load_cfg()

        row_dataset_view = pn.Row(self.dataset_package, pn.pane.HTML('<span> </span>'), self.master_dataset_package,
                margin=(5, 0), sizing_mode='stretch_width')

        self.bt_compute = pn.widgets.Button(name='Compute', icon='sum', button_type='primary', align=('start', 'center'), width=bt_size, description='Compute the crossdating.')
        self.bt_compute.on_click(self.on_compute)
        self.bt_stop = pn.widgets.Button(name='Stop', icon='stop', button_type='primary', align=('start', 'center'), width=bt_size, description='Stop the computation.')
        self.bt_stop.on_click(self.on_stop)

        self.progress = pn.indicators.Progress(name='Run', value=0, width=250, disabled=True, bar_color='primary')
        self.progress_info = pn.pane.HTML()
        self.cross_dating.progress.param.watch(self.on_progress, ['count'], onlychanged=True)

        self.bt_sync = pn.widgets.Button(name='Sync / Unsync', icon='swicth', button_type='primary', align=('start', 'center'), width=bt_size, description='Switch sync/unsync the seleted rows.')
        self.bt_sync.on_click(self.on_sync)
        
        self.bt_export = pn.widgets.FileDownload(callback=self.on_export, filename='crossdating.xlsx', embed=False, 
                                                label='Download', icon='file-export', button_type='primary', 
                                                width=bt_size, align=('start', 'center'), description='Download the table as an Excel file.')


        self.param_array.param.watch(self.on_group_by, ['group_by'], onlychanged=True)
        self.param_array_columns.columns.param.watch(self.on_columns, ['value'], onlychanged=True)
        self.param_array.param.watch(self.on_table_height, ['table_height'], onlychanged=True)

        self.param_filter.param.watch(self.on_tabs, ['score', 'filter_threshold', 'threshold', 'filter_max_rank', 'max_rank', 'filter_sync', 'sync', 'no_dates_in_the_future'], onlychanged=True)
        
        self.cross_dating.param_matrix.param.watch(self.on_tabs, ['size_scale', 'font_scale',  'method', 'metric', 'sorted', 'color_map'], onlychanged=True)
        self.cross_dating.param_stem.param.watch(self.on_tabs, ['keycode_nrows', 'height', 'window_size'], onlychanged=True)
        self.cross_dating.param_density.param.watch(self.on_tabs, ['bullet_size', 'font_size', 'keycode_nrows', 'method', 'bins_size'], onlychanged=True)
        self.cross_dating.param_density.param.watch(self.on_density_width, ['width'], onlychanged=True)
        self.cross_dating.param_density.param.watch(self.on_density_height, ['height'], onlychanged=True)
        self.cross_dating.param_graph.param.watch(self.on_tabs, [ 'height', 'width', 'font_size', 'line_ratio', 'bullet_ratio'], onlychanged=True)

        self.cross_dating.param_map.param.watch(self.on_tabs, ['width', 'map_provider', 'label_distance', 'alpha', 'map_type', 'map_radius', 'map_center','line_ratio', 'height', 'font_size', 'bullet_ratio'], onlychanged=True)

        self.wtabulator = self.get_tabulator()
        self.run_thread = None
        self.wrun = pn.Row(
                        self.bt_compute,
                        pn.Column(
                            self.progress_info,
                            self.progress,
                            ),
                        self.bt_stop,
                        self.bt_sync, 
                        self.bt_export
                    )

        self.warray = pn.Column(#pn.Row(self.bt_sync, self.bt_export), 
                                self.wrun,
                                self.wtabulator)
        self.wheat_matrix = pn.pane.Matplotlib()
        self.wstem = pn.pane.Bokeh()
        self.wzstem = pn.pane.Bokeh()
        self.wdensity = pn.Column()
        self.wmap = pn.pane.Bokeh()
        self.wgraph = pn.pane.Bokeh()
        self.plot = pn.Row()
        
        self.tabs = pn.Tabs(('Array', pn.Row(self.warray, self.plot)),
                           ('Matrix', pn.Column('Computed only on sync *scores*. Threshold filter allowed.', self.wheat_matrix)), 
                           ('Timeline', pn.Column('Computed on all *scores*. Threshold and rank filters allowed.', self.wstem, self.wzstem, sizing_mode='stretch_width')), 
                           ('Density', pn.Column('Computed on all *scores*. Threshold and rank are disallowed.', self.wdensity)),
                           ('Map', pn.Column('Computed only on sync *scores*. Threshold filter allowed.', self.wmap)),
                           ('Graph', pn.Column('Computed only on sync *scores*. Threshold filter allowed.', self.wgraph)),
                           dynamic=False)
        self.tabs.param.watch(self.on_tabs,  ['active'], onlychanged=True)
        
        self._layout = pn.Column(
                row_dataset_view, 
                #self.wrun,
                self.tabs,
                name=self.name,
                margin=(5, 0), sizing_mode='stretch_width')

    def on_table_height(self, event):
        self.wtabulator.max_height = event.new
        self.wtabulator.min_height = event.new

    def get_tabulator(self):
        """
        Returns a Tabulator widget populated with data.

        Returns:
            pn.widgets.Tabulator: A Tabulator widget.
        """
        
        cols = self.param_array_columns
        dtype = cols.get_dtype()
        return pn.widgets.Tabulator(pd.DataFrame(columns=cols.get_columns_hiddens()),
                                    hidden_columns=cols.get_hiddens(),  
                                    text_align=_cell_text_align(dtype),
                                    editors=_cell_editors(dtype, False),
                                    header_filters=_header_filters_crossdating(dtype),
                                    formatters=_cell_formatters(dtype, bool_formater=True),
                                    pagination='remote',
                                    page_size=10000,
                                    frozen_columns=[KEYCODE, KEYCODE_MASTER],
                                    selectable='checkbox', 
                                    sizing_mode='stretch_width',
                                    layout='fit_data_fill',
                                    height_policy='max',
                                    max_height=self.param_array.table_height,
                                    min_height=self.param_array.table_height,
                                    show_index = False,
                                    margin=(0,0),
                                    #row_content = self.get_plot,
                                    )

    
    def _sync_data(self, event):
        self.clean()

    # def get_plot(self, row):
    #     """
    #     Generates a plot based on the given row data.

    #     Parameters:
    #         row (pandas.Series): The row data containing information for generating the plot.

    #     Returns:
    #         pn.pane.Bokeh: The generated plot as a Bokeh pane.

    #     Raises:
    #         None

    #     """
    #     def callback(event):
    #         nonlocal data, data_dt, data_log, data_dt_master   
    #         if data is not None:
    #             x = int(round(event.x)) 
    #             pos = np.searchsorted(ds.data['x'], x)
    #             if event.modifiers['shift']:
    #                 if ds.data['y'][pos] is np.nan:
    #                     #perror('callback shift nan', event.x, event.y, event.modifiers)   
    #                     ds.data['x'] = np.delete(ds.data['x'], -1)
    #                     ds.data['y'] = np.delete(ds.data['y'], pos)
    #                     data = np.delete(data, pos)
    #                     data_dt = np.delete(data_dt, pos)
    #                     data_log = np.delete(data_log, pos)     
    #                 else:
    #                     logger.warning('callback shift not nan', event.x, event.y, event.modifiers)
    #             else:                
    #                 if data is not None:
    #                     ds.data['x'] = np.append(ds.data['x'], ds.data['x'][-1] + 1) 
    #                     ds.data['y'] = np.insert(ds.data['y'], pos, np.nan) 
    #                     data = np.insert(data, pos, np.nan) 
    #                     data_dt = np.insert(data_dt, pos, np.nan) 
    #                     data_log = np.insert(data_log, pos, np.nan) 
                        
    #             start_data = max(0, -offset)
    #             start_data_master = max(0, offset)
    #             l = min(len(data_dt) - start_data, len(data_dt_master) - start_data_master)
    #             segment = data_dt[start_data:start_data+l]
    #             master_segment = data_dt_master[start_data_master:start_data_master+l]
    #             valid_data_points = (~np.isnan(segment)) & (~np.isnan(master_segment))
    #             n = np.sum(valid_data_points)
    #             nnan =  valid_data_points.size - n
    #             nr, nt, ___ = r(n, segment[valid_data_points], master_segment[valid_data_points])
    #             wr.value = round(nr, 3)
    #             wt.value = round(nt, 3)
    #             perror('callback', nr,nt, event.x, event.y, event.modifiers)
        
    #     def get_data(data_type, data, data_dt, data_log):
    #         if data_type == 'Raw':
    #             return data
    #         elif data_type == 'Log':
    #             return data_log
    #         else:
    #             return data_dt
        
    #     def on_data_type(event):
    #         if ds.data is not None:
    #             ds.data['y'] = get_data(event.new, data, data_dt, data_log)
    #             ds_master.data['y'] = get_data(event.new, data_master, data_dt_master, data_log_master)
            
    #     plot = None   
    #     ds = ColumnDataSource()
    #     ds_master = ColumnDataSource()
    #     try:
    #         wdata_type = pn.widgets.RadioBoxGroup(name='Data Type', value='Raw', options=['Raw', 'Log', 'Detrend'], inline=True)
    #         wdata_type.param.watch(on_data_type, ['value'],  onlychanged=True)
    #         wr = pn.widgets.StaticText(name='r (new)', value='-')
    #         wt = pn.widgets.StaticText(name='t-score (new)', value='')
    #         data, data_dt, data_log = self.dataset_package.get_data_values_id(row[ID])
            
    #         dp = self.dataset_package if self.inner else self.master_dataset_package
    #         data_master, data_dt_master, data_log_master = dp.get_data_values_id(row[ID_MASTER])

    #         offset = row[OFFSET]

    #         ds.data['x'] =  np.arange(len(data)) + offset
    #         ds.data['y'] =  get_data(wdata_type.value, data, data_dt, data_log)
    #         #perror('get_plot', ds.data['y'])
    #         ds_master.data['x'] = np.arange(len(data_master))
    #         ds_master.data['y'] = get_data(wdata_type.value, data_master, data_dt_master, data_log_master)
    #         #perror('get_plot', ds_master.data['y'])
            
    #         fig = figure(margin=(5), height=200, width=1000,
    #                 tools="pan,wheel_zoom,box_zoom,reset,hover,crosshair,tap,save", 
    #                 toolbar_location="left",
    #                 tooltips=[('(date/offset,value)', '(@x, @y)')],
    #                 sizing_mode='stretch_width')
    #         fig.line('x', 'y', source=ds, line_width=1, line_color=Category10[10][0], legend_label='serie')
    #         fig.line('x', 'y', source=ds_master, line_width=1, line_color=Category10[10][1], legend_label='master')
            
    #         fig.on_event(DoubleTap, callback)
    #         plot = pn.pane.Bokeh(fig, sizing_mode='stretch_width')
    #         return pn.Column(pn.Row(wdata_type, wr, wt), plot, height=self.param_array.table_height//2)
    #     except Exception:
    #         logger.error(f'get_plot: {inst}', exc_info=True)
    #         plot = None
    #         return pn.Column()
            
    def on_export(self):
        try:
            if (self.wtabulator.value is None) or (len(self.wtabulator.value) <=0 ):
                logger.warning('No data')
                return     
            output = BytesIO()
            chunk_size = 1000000
            with pd.ExcelWriter(output) as writer:
                data = self.row_results if self.row_results is not None else self.wtabulator.value
                total_rows = len(data)
                num_chunks = (total_rows // chunk_size) + 1
                
                for i in range(num_chunks):
                    start_row = i * chunk_size
                    end_row = min((i + 1) * chunk_size, total_rows)
                    chunk = data.iloc[start_row:end_row]
                    sheet_name = f'crossdating_part_{i + 1}'
                    chunk.to_excel(writer, sheet_name=sheet_name, merge_cells=False, float_format="%.6f")
            
            output.seek(0)
            return output

        except Exception as inst:
            logger.error(f'on_excel: {inst}', exc_info=True)

    def on_columns(self, event):
        self.param_array_columns.update_tabulator(self.wtabulator)

    def on_group_by(self, event):
        if self.param_array.group_by is None:
            self.wtabulator.groupby = []
        else:
            if self.param_array.group_by == KEYCODE+'/'+KEYCODE_MASTER:
                self.wtabulator.groupby = [KEYCODE, KEYCODE_MASTER]
            else:
                self.wtabulator.groupby = [self.param_array.group_by]

    def on_density_height(self, event):
        try:
            self._layout.loading = True
            for p in self.wdensity.objects:
                p.height = self.cross_dating.param_density.height
        except Exception as inst:
            logger.error(f'Crossdating: {inst}', exc_info=True)
            self.wdensity.objects = []
        finally:
            self._layout.loading = False

    def on_density_width(self, event):
        try:
            self._layout.loading = True
            for p in self.wdensity.objects:
                p.width = self.cross_dating.param_density.width
        except Exception as inst:
            logger.error(f'Crossdating: {inst}', exc_info=True)
            self.wdensity.objects = []
        finally:
            self._layout.loading = False

    def get_score_key(self):
        if self.param_filter.score == CORR and CORRELATION in self.cross_dating.method:
            return CORR
        elif self.param_filter.score == T_SCORE and CORRELATION in self.cross_dating.method:
            return T_SCORE
        elif self.param_filter.score == GLK and GLK in self.cross_dating.method:
            return GLK
        elif self.param_filter.score == Z_SCORE and GLK in self.cross_dating.method:
            return Z_SCORE
        elif self.param_filter.score == DISTANCE and DISTANCE in self.cross_dating.method:
            return DISTANCE
        else:
            logger.warning(f'score key {self.param_filter.score} not in method ({self.cross_dating.method}), score set to {self.cross_dating.method[0]}')
            if len(self.cross_dating.method) > 0:
                self.param_filter.score = self.cross_dating.method[0]
                return self.param_filter.score
        return None

    def on_tabs(self, event):
        def get_filter(threshold, rank, sync, unk_date):
            param = {}
            param['score'] = self.get_score_key()

            if self.param_filter.filter_threshold and threshold:
                param['threshold'] = self.param_filter.threshold
            if self.param_filter.filter_max_rank and rank:
                param['max_rank'] = self.param_filter.max_rank
            if self.param_filter.filter_sync and sync:
                param['sync'] = self.param_filter.sync
            if self.param_filter.no_dates_in_the_future and unk_date:
                param['incorrect_date'] = self.param_filter.no_dates_in_the_future
                
            return param
        
        def concat_data():
            dt_data = self.dataset_package.data
            dt_data_master = self.master_dataset_package.data 
            if self.inner is None:
                return dt_data
            else:
                if (dt_data is None) and (dt_data_master is None):
                    return None
                elif self.cross_dating.param_map.map_type == 0:
                    return dt_data_master
                else:
                    with warnings.catch_warnings():
                        # TODO: pandas 2.1.0 has a FutureWarning for concatenating DataFrames with Null entries
                        warnings.filterwarnings("ignore", category=FutureWarning)
                        return pd.concat([dt_data, dt_data_master ], ignore_index=True)

        try:
            self._layout.loading = True
            if self.tabs.active == 0: #Array
                self.row_results = self.cross_dating.concat_results(**get_filter(True, True, True, True))
                if self.row_results is None:
                    df = pd.DataFrame(columns=self.param_array_columns.get_columns())
                elif len(self.row_results) < self.param_array.max_results:
                    df = self.row_results
                else:
                    logging.warning(f'Too many results, the displayed array will be limited to {self.param_array.max_results/1000000}M of values.')
                    df = self.row_results.sort_values(by=self.get_score_key()).iloc[ :self.param_array.max_results]
                self.set_tabulator_value(df)
            elif self.row_results is not None:
                if self.tabs.active == 1: #Matrix
                    self.wheat_matrix.object = self.cross_dating.heat_matrix(**get_filter(True, False, False, False), 
                                                metric=self.cross_dating.param_matrix.metric, method=self.cross_dating.param_matrix.method)
                elif self.tabs.active == 2: #Time line
                    self.wstem.object, self.wzstem.object = self.cross_dating.stem(**get_filter(True, True, False, False))
                elif self.tabs.active == 3: 
                    rows, _ = self.cross_dating.density(score=self.get_score_key())
                    self.wdensity.objects = rows
                elif self.tabs.active == 4:
                    try:
                        self.wmap.object = self.cross_dating.map( **get_filter(True, True, False, False), data_dt=concat_data())
                    except MapValueError as inst:
                        logger.warning(f'CrossDating: {repr(inst)}')
                elif self.tabs.active == 5: 
                    self.wgraph.object = self.cross_dating.graph( **get_filter(True, True, False, False))
        except Exception as inst:
            logger.error(f'Crossdating: {inst}', exc_info=True)
            self.set_tabulator_value(pd.DataFrame(columns=self.param_array_columns.get_columns()))
            self.wheat_matrix.object = None
            self.wstem.object = None
            self.wdensity.objects = []
        finally:
            self._layout.loading = False

    def set_tabulator_value(self, value):
        self.wtabulator.value = value
        self.on_group_by(None)

    @param.depends("inner", watch=True)
    def _update_inner(self):
        if self.inner == True :
            self.master_dataset_package._layout.visible = False
        else:
            self.master_dataset_package._layout.visible = True

    def get_sidebar(self, visible=True):
        self.p_panel = pn.Param(self.param, show_name=False)
        self.p_filter = pn.Param(self.param_filter, show_name=True)
        self.p_cross = pn.Param(self.cross_dating, show_name=False)
        self.parray = pn.Column(*self.param_array_columns.get_sidebar().objects, pn.Param(self.param_array, show_name=False), name='Array')
        self.pmatrix = pn.Param(self.cross_dating.param_matrix, show_name=False)
        self.pstem = pn.Param(self.cross_dating.param_stem, show_name=False)
        self.phist = pn.Param(self.cross_dating.param_density, show_name=False)
        self.pmap = pn.Param(self.cross_dating.param_map, show_name=False)
        self.pgraph = pn.Param(self.cross_dating.param_graph, show_name=False)
        
        return pn.Card(self.p_panel, self.p_cross, self.p_filter, 
                       pn.Accordion(self.parray, self.pmatrix, self.pstem, self.phist, self.pmap, self.pgraph, toggle=True),
                title='Crossdating', sizing_mode='stretch_width', margin=(5, 0), collapsed=True, visible=visible)  

    def dump_cfg(self):
        with open(self.cfg_file, 'w') as fd:
            data = {
                'version': __version__,
                'param' : self.param.serialize_parameters(),
                'param_filter' : self.param_filter.param.serialize_parameters(),
                'param_array' : self.param_array.param.serialize_parameters(),
                'cross_dating' : self.cross_dating.param.serialize_parameters(),
                'cross_dating.param_matrix' : self.cross_dating.param_matrix.param.serialize_parameters(),
                'cross_dating.param_stem' : self.cross_dating.param_stem.param.serialize_parameters(),
                'cross_dating.param_density' : self.cross_dating.param_density.param.serialize_parameters(),
                'cross_dating.param_map' : self.cross_dating.param_map.param.serialize_parameters(),
                'cross_dating.param_graph' : self.cross_dating.param_graph.param.serialize_parameters(),
            }
            json.dump(data, fd)

    def load_cfg(self):
        cross_dating = CrossDating()
        try:
            if check_version(self.cfg_file):
                with open(self.cfg_file, 'r') as fd:
                    data = json.load(fd)
                    self.param_filter = self.ParamFilter(**self.ParamFilter.param.deserialize_parameters(data['param_filter']))
                    self.param_array = self.ParamArray(**self.ParamArray.param.deserialize_parameters(data['param_array']))
                    cross_dating = CrossDating(**CrossDating.param.deserialize_parameters(data['cross_dating']))
                    cross_dating.param_matrix = CrossDating.ParamMatrix(** CrossDating.ParamMatrix.param.deserialize_parameters(data['cross_dating.param_matrix']))
                    cross_dating.param_stem =  CrossDating.ParamStem(** CrossDating.ParamStem.param.deserialize_parameters(data['cross_dating.param_stem']))
                    cross_dating.param_density =  CrossDating.ParamDensity(** CrossDating.ParamDensity.param.deserialize_parameters(data['cross_dating.param_density']))
                    cross_dating.param_map =  CrossDating.ParamMap(** CrossDating.ParamMap.param.deserialize_parameters(data['cross_dating.param_map']))
                    cross_dating.param_graph =  CrossDating.ParamGraph(** CrossDating.ParamGraph.param.deserialize_parameters(data['cross_dating.param_graph']))
                    (json.loads(data['param']))
                    for key, value in json.loads(data['param']).items():
                        if key in self.param.params().keys():
                            if key != 'name':
                                self.param.set_param(key, value)
        except Exception as inst:
            logger.warning(f'ignore {self.cfg_file} {inst}.')
        finally:
            return cross_dating

    def columns(self):
        return list(CrossDating.COLS.keys())

    def dtype_columns(self):
        return CrossDating.COLS
        
    def __panel__(self):
        return self._layout

    def get_selection(self) -> pd.DataFrame:
        """
        Returns the view of selectionned series. 
        """
        return _get_selection(self.wtabulator)
    
    def on_link(self, event):
        if not self.inner:
            raise ValueError('Only available for self crossdating (inner parameter must be True)')
        selections = self.get_selection()
        logger.warning('link selection is not implemented yet')
        
    def on_sync(self, event):
        if self.cross_dating.df_results is None:
            logger.warning('No data.')
            return
        
        self._layout.loading = True
        try:
            cp = copy.deepcopy(self.wtabulator.value)
            selections = self.get_selection()        
            unique_pairs = selections[[ID, ID_MASTER]].drop_duplicates().apply(tuple, axis=1).tolist()
            df = self.cross_dating.df_results
            filtered_df = df[df.apply(lambda row: (row[ID], row[ID_MASTER]) in unique_pairs, axis=1)]
            sync_index = []
            sync_date = []
            for (id, id_master), grp in filtered_df.groupby([ID, ID_MASTER]):
                if grp[SYNC].sum() > 1:
                    logger.warning(f'Multipled sync pairs are ignored in {grp.iat[0, [KEYCODE, KEYCODE_MASTER, SYNC]]}.')
                else:
                    df.loc[grp.index, SYNC] = False
                    idx = grp.index.intersection(selections.index)                
                    df.loc[idx, SYNC] = True
                    sync_index += idx.tolist()
                    sync_date += df.loc[idx, DATE_BEGIN_ESTIMATED].tolist()
        
            self.row_results[SYNC] = df.loc[self.row_results.index, SYNC]
            self.wtabulator.patch({SYNC: list(self.row_results[SYNC].items())})
            cp = copy.deepcopy(self.wtabulator.value)
            for i, date_begin in zip(sync_index, sync_date):
                id = df.at[i, ID]
                self.dataset_package.dataset.set_inconsistent_on_ascendants(id, notify=False) 
                self.dataset_package.dataset.set_dates(id, date_begin, warning=False)
                self.dataset_package.dataset.log_crossdating(df.loc[id,:].to_dict())
        except Exception as inst:
            logger.error(f'sync/unsync: {inst}.')
        finally:
            self.set_tabulator_value(cp)
            self._layout.loading = False

    def on_stop(self, event):
        if self.run_thread is not None:
            self.run_thread.stop()

    def clean(self):
        self.plot = None
        self.warray[-1] = self.wtabulator
        self.wheat_matrix.object = None
        self.wstem.object = None
        self.wzstem.object = None
        self.wdensity.object = None
        self.wmap.object = None
        self.wgraph.object = None

    def on_compute(self, event):
        if self.dataset_package.dt_data is None:
            logger.warning('No data to process.')
            return
        try:
            dt_data = self.dataset_package.dt_data
            dt_data_master = self.master_dataset_package.dt_data if not self.inner else None
            if (not self.inner) and (self.dataset_package.dt_param != self.master_dataset_package.dt_param):
                raise ValueError('master and dataset parameters are not equal')
            
            self.progress_info.object = f'<span>start running...</span>'
            self.run_thread = RunThread(run_action=self.cross_dating.run, end_action=self.on_end, args=(self.dataset_package.get_data_type(), dt_data, dt_data_master))
            self.run_thread.start()
            
        except Exception as inst:
            logger.error(f'on_compute: {inst}', exc_info=True)
            self.clean()

    def on_end(self):
        
        name = 'crossdating_' + self.dataset_package.get_package_name() 
        if not self.inner:
            name += '_' + self.master_dataset_package.get_package_name()
        self.bt_export.filename = name + '.xlsx'

        self.progress.disabled = True
        rate, info = self.cross_dating.progress.info() 
        self.progress.value = rate
        self.progress_info.object = f'<span>{info}</span>'
        self.run_thread = None
        self.on_tabs(None)        

    def on_progress(self, event):
        if self.cross_dating.progress.count == self.cross_dating.progress.max_count:
            self.progress.disabled = True
            return
        self.progress.disabled = False
        rate, info = self.cross_dating.progress.info() 
        self.progress.value = rate
        self.progress_info.object = f'<span>{info}</span>'
        